from datetime import datetime

from pytz import timezone

from .abstract_rule import ValidationRule
from ..exceptions import ValidationException


class DatesOverlapRule(ValidationRule):
    VALID_TIMEZONES = ['Pacific/Midway', 'Pacific/Honolulu', 'America/Anchorage', 'America/Los_Angeles',
                       'America/Denver', 'America/Chicago', 'America/New_York', 'America/Halifax',
                       'America/St_Johns', 'America/Sao_Paulo', 'Atlantic/South_Georgia', 'Atlantic/Cape_Verde',
                       'Etc/GMT', 'Europe/London', 'Europe/Paris', 'Europe/Kaliningrad', 'Europe/Moscow',
                       'Asia/Tehran', 'Asia/Dubai', 'Asia/Kabul', 'Asia/Karachi', 'Asia/Kolkata', 'Asia/Kathmandu',
                       'Asia/Dhaka', 'Asia/Bangkok', 'Asia/Shanghai', 'Asia/Tokyo', 'Australia/Adelaide',
                       'Australia/Sydney', 'Pacific/Pohnpei', 'Pacific/Auckland']

    def __init__(self, start_date, end_date, zone_name):

        self.start_date = start_date
        self.end_date = end_date

        self.zone_name = zone_name

    def execute(self):
        """
        rules:

        start_date is later than end_date. Message "start_date must be earlier than end_date"
        start_time is later than end_time. Message " start_time must be earlier than end_time"
        start_date is in the past. Message "start_date must be in the future"

        :return:
        """
        # validate timezone string
        if self.zone_name not in self.VALID_TIMEZONES:
            raise ValidationException("zone_name must be one of correct timezone names, e.g. 'America/New_York'")

        # combine dates
        tz = timezone(self.zone_name)
        try:
            self.start_date = datetime.strptime(self.start_date, "%Y-%m-%dT%H:%M")
            self.end_date = datetime.strptime(self.end_date, "%Y-%m-%dT%H:%M")
        except ValueError:
            raise ValidationException("dates must be provided in %Y-%m-%dT%H:%M format")

        start_date = datetime.combine(self.start_date, datetime.min.time())
        start_date = start_date.replace(hour=self.start_date.hour, minute=self.start_date.minute)
        start_date = tz.localize(start_date)

        end_date = datetime.combine(self.end_date, datetime.min.time())
        end_date = end_date.replace(hour=self.end_date.hour, minute=self.end_date.minute)
        end_date = tz.localize(end_date)

        # start_time is later than end_time
        if start_date >= end_date:
            # case: for the same date but time, "start_time must be earlier than end_time"
            if start_date.date() == end_date.date():
                raise ValidationException("start_time must be earlier than end_time")
            else:
                raise ValidationException("start_date must be earlier than end_date")

        if start_date.date() < datetime.now(tz=tz).date():
            raise ValidationException("start_date must be in future")



import inspect
from jsonschema import ValidationError


class ValidationException(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)

        stack = inspect.stack()

        self.class_caller = stack[1][1]
        self.class_line = stack[1][2]


class JsonSchemaException(Exception):
    def __init__(self, ex):
        assert isinstance(ex, ValidationError)
        message = str(ex)

        # we need to replace standard errors with custom
        attr = "" if len(ex.path) < 1 else ex.path[0]
        enum_text = "value must be one of the following values"
        if ex.validator in ["type", "enum", "format", "required"]:
            if ex.validator_value == "boolean":
                message = "%s value must be a boolean" % attr
            elif ex.validator_value == "integer":
                message = "%s value must be an integer" % attr
            elif ex.validator_value == "number":
                message = "%s value must be a float" % attr
            elif ex.validator == "enum":
                message = "%s value must be one of the following values ('%s')" % (attr, "','".join(ex.validator_value))
            elif ex.validator_value == "date":
                message = "%s value must be a date in YYYY-MM-DD format" % attr
            elif ex.validator_value == "time":
                message = "%s value must be a time in HH:MM:SS format" % attr
            elif ex.validator_value == "string":
                message = "%s value must be a string" % attr
            elif ex.validator == "type" and ex.validator_value == "array":
                message = "%s should have valid format, positive numbers separated by comma" % attr
            elif ex.validator == "required":
                message = str(ex.message)
            elif ex.validator_value == "email":
                message = "'{}' should contain valid email address".format(attr)
        elif ex.validator == "minimum":
            message = "{field_name} must be greater than {min_value}".format(field_name=attr,
                                                                             min_value=ex.validator_value)
        elif ex.validator == "minLength":
            # definitely we need to check for non-empty values
            if len(ex.instance) < 1:
                message = "{} is a mandatory field".format(attr)
            else:
                message = "{field_name} is too short, must be from {min_value} to {max_value} characters long".format(field_name=attr,
                                                                                                                      min_value=ex.schema.get("minLength"),
                                                                                                                      max_value=ex.schema.get("maxLength"))

        super().__init__(message)



from enum import unique

from ts_t1_validator.models.enums.abstract_enum import AbstractEnum


@unique
class NotificationTypeEnum(AbstractEnum):
    CAMPAIGN = "campaign"
    ORGANIZATION = "organization"
    AGENCY = "agency"
    ADVERTISER = "advertiser"
    UNDEFINED = None



from typing import Dict

from ts_t1_validator.services.t1 import T1Service
from ts_t1_validator.validators.abstract_validator import AbstractValidator
from ts_t1_validator.validators.campaign_post_validator import CampaignPostValidator
from ts_t1_validator.validators.strategy_post_validator import StrategyPostValidator


class T1Validator:
    def __init__(self, t1_creds: Dict, validator_type="campaign"):
        self.validator = None
        t1_service = T1Service(t1_creds=t1_creds)
        if validator_type == "campaign":
            self.validator = CampaignPostValidator(t1_service=t1_service)
        elif validator_type == "strategy":
            self.validator = StrategyPostValidator(t1_service=t1_service)
        else:
            raise NotImplemented()

    def validate(self, *args, **kwargs):
        """
        Proxy validate method to proper class
        :return:
        """
        return self.validator.validate(*args, **kwargs)



from ts_t1_validator.utils import is_number
from ts_t1_validator.validators.rules.abstract_rule import ValidationRule
from ts_t1_validator.validators.rules.frequency_type import FrequencyTypeRule
from ts_t1_validator.models.enums.frequency_interval import FrequencyIntervalEnum
from ts_t1_validator.models.enums.frequency_type import FrequencyTypeEnum


class FrequencyTypePutRule(ValidationRule):
    def __init__(self, frequency_optimization: bool,
                 frequency_type: FrequencyTypeEnum,
                 frequency_amount,
                 frequency_interval: FrequencyIntervalEnum):

        assert type(frequency_type) is FrequencyTypeEnum, "frequency_type should be type of FrequencyTypeEnum"
        assert type(frequency_interval) is FrequencyIntervalEnum, \
            "frequency_interval should be type of FrequencyIntervalEnum"

        self.frequency_optimization = frequency_optimization
        self.frequency_type = frequency_type
        self.frequency_amount = frequency_amount
        self.frequency_interval = frequency_interval

    def put(self, frequency_type: FrequencyTypeEnum,
            frequency_interval: FrequencyIntervalEnum,
            frequency_amount: int or float or None):
        """
        :param frequency_type: FrequencyTypeEnum
        :param frequency_interval: FrequencyIntervalEnum
        :param frequency_amount: int or float or None
        :return: self
        """

        if self.frequency_optimization:
            self.frequency_amount = frequency_amount
            self.frequency_type = frequency_type
            self.frequency_interval = frequency_interval

        if self.frequency_type is FrequencyTypeEnum.NO_LIMIT:
            self.frequency_interval = frequency_interval
            self.frequency_amount = frequency_amount

        if is_number(frequency_amount):
            self.frequency_amount = frequency_amount

        return self

    def execute(self):
        """
        rule:
        frequency_optimization, frequency_type, frequency_amount, frequency_interval
            - if undefined in DB, follow POST logic
        """

        validator = FrequencyTypeRule(
            frequency_optimization=self.frequency_optimization,
            frequency_type=self.frequency_type,
            frequency_amount=self.frequency_amount,
            frequency_interval=self.frequency_interval)

        return validator.execute()



from ts_t1_validator.validators.rules.abstract_rule import ValidationRule
from ts_t1_validator.validators.rules.strategy_required_fields_post import StrategyPostRequiredFieldsRule
from ts_t1_validator.models.enums.impression_pacing_type import ImpressionPacingTypeEnum
from ts_t1_validator.models.enums.pacing_type import PacingTypeEnum


class StrategyPutRequiredFieldsRule(ValidationRule):
    def __init__(self, pacing_type, impression_pacing_type, min_bid,
                 max_bid, goal_value, roi_target, frequency_optimization):
        assert type(pacing_type) is dict, "pacing_type should be dictionary type"
        assert type(pacing_type.get("db")) is PacingTypeEnum, "pacing_type['db'] should be PacingTypeEnum type"
        assert type(pacing_type.get("dto")) is PacingTypeEnum, "pacing_type['dto'] should be PacingTypeEnum type"

        assert type(impression_pacing_type) is dict, "impression_pacing_type should be dictionary type"
        assert type(impression_pacing_type.get(
            "db")) is ImpressionPacingTypeEnum, "impression_pacing_type['db'] should be ImpressionPacingTypeEnum type"
        assert type(impression_pacing_type.get(
            "dto")) is ImpressionPacingTypeEnum, "impression_pacing_type['dto'] should be ImpressionPacingTypeEnum type"

        assert type(min_bid) is dict, "min_bid should be dictionary type"
        assert type(max_bid) is dict, "max_bid should be dictionary type"
        assert type(goal_value) is dict, "goal_value should be dictionary type"
        assert type(roi_target) is dict, "roi_target should be dictionary type"
        assert type(frequency_optimization) is dict, "frequency_optimization should be dictionary type"

        self.pacing_type = pacing_type
        self.impression_pacing_type = impression_pacing_type
        self.min_bid = min_bid
        self.max_bid = max_bid
        self.goal_value = goal_value
        self.roi_target = roi_target
        self.frequency_optimization = frequency_optimization

    def execute(self):
        """
        check for required fields step by step
        :return:
        """

        """
        rule
            pacing_type, impression_pacing_type, min_bid, max_bid, goal_value, roi_target, frequency_optimization
            - if undefined in DB, follow POST logic
        """
        if self.__oneOfRequiredFieldUndefinedInDb():
            postValidation = StrategyPostRequiredFieldsRule(pacing_type=self.getUpdatedValue("pacing_type"),
                                                            impression_pacing_type=self.getUpdatedValue(
                                                                "impression_pacing_type"),
                                                            min_bid=self.getUpdatedValue("min_bid"),
                                                            max_bid=self.getUpdatedValue("max_bid"),
                                                            goal_value=self.getUpdatedValue("goal_value"),
                                                            roi_target=self.getUpdatedValue("roi_target"),
                                                            frequency_optimization=self.getUpdatedValue(
                                                                "frequency_optimization"))
            return postValidation.execute()

    def __oneOfRequiredFieldUndefinedInDb(self):
        """
        check that one of required fields is undefined
        :return:
        """
        return self.pacing_type["db"] is PacingTypeEnum.UNDEFINED or self.impression_pacing_type[
            "db"] is ImpressionPacingTypeEnum.UNDEFINED or self.min_bid["db"] is None or self.max_bid[
            "db"] or self.goal_value["db"] is None or self.roi_target["db"] is None or self.frequency_optimization["db"] is None



from typing import Optional, Union

from ts_t1_validator.utils import is_number, is_int
from ts_t1_validator.validators.exceptions import ValidationException
from ts_t1_validator.validators.rules.abstract_rule import ValidationRule
from ts_t1_validator.models.enums.impression_pacing_type import ImpressionPacingTypeEnum
from ts_t1_validator.models.enums.pacing_interval import PacingIntervalEnum


class ImpressionPacingTypeRule(ValidationRule):
    def __init__(self,
                 impression_pacing_type: ImpressionPacingTypeEnum,
                 impression_pacing_amount: Optional[Union[int, float]],
                 impression_pacing_interval: PacingIntervalEnum):
        self.impression_pacing_type = impression_pacing_type
        self.impression_pacing_amount = impression_pacing_amount
        self.impression_pacing_interval = impression_pacing_interval

    def execute(self):
        if self.impression_pacing_type is not ImpressionPacingTypeEnum.UNDEFINED:
            """
            rule:
                impression_pacing_amount required if impression_pacing_type is asap or even
                    - error message "{:field_name} is required when impression_pacing_type is asap or even"
                must be undefined if impression_pacing_type is no-limit
                    - error message "{:field_name} cannot be defined when impression_pacing_type is no-limit"
            """
            if self.impression_pacing_type in [ImpressionPacingTypeEnum.ASAP,
                                               ImpressionPacingTypeEnum.EVEN] and self.impression_pacing_amount is None:
                raise ValidationException("impression_pacing_amount is required when impression_pacing_type is asap or even")
            elif self.impression_pacing_type is ImpressionPacingTypeEnum.NO_LIMIT and self.impression_pacing_amount is not None:
                raise ValidationException("impression_pacing_amount cannot be defined when impression_pacing_type is no-limit")

            """
            rule:
                impression_pacing_interval required if impression_pacing_type is asap or even
                    - error message "{:field_name} is required when impression_pacing_type is asap or even"
                must be undefined if impression_pacing_type is no-limit
                    - error message "{:field_name} cannot be defined when impression_pacing_type is no-limit
            """
            if self.impression_pacing_type in [ImpressionPacingTypeEnum.ASAP,
                                               ImpressionPacingTypeEnum.EVEN] and self.impression_pacing_interval is PacingIntervalEnum.UNDEFINED:
                raise ValidationException("impression_pacing_interval is required when impression_pacing_type is asap or even")
            elif self.impression_pacing_type is ImpressionPacingTypeEnum.NO_LIMIT and self.impression_pacing_interval is not PacingIntervalEnum.UNDEFINED:
                raise ValidationException("impression_pacing_interval cannot be defined when impression_pacing_type is no-limit")

            """
            rule: for impression_pacing_amount, while input_type='fixed', type must be int
            rule: for impression_pacing_amount, while input_type='percentage', type must be float
            """
            if self.impression_pacing_type in [ImpressionPacingTypeEnum.ASAP,
                                               ImpressionPacingTypeEnum.EVEN]:
                if not is_number(self.impression_pacing_amount):
                    raise ValidationException("impression_pacing_amount must be a float")
        else:
            if self.impression_pacing_interval is not PacingIntervalEnum.UNDEFINED:
                raise ValidationException("impression_pacing_type is required when impression_pacing_interval is defined")
            if self.impression_pacing_amount is not None:
                raise ValidationException("impression_pacing_interval is required when impression_pacing_amount is defined")



from typing import Optional, Union

from ts_t1_validator.models.enums.goal_type import GoalTypeEnum
from ts_t1_validator.utils import is_number
from ts_t1_validator.validators.exceptions import ValidationException
from ts_t1_validator.validators.rules.abstract_rule import ValidationRule
from ts_t1_validator.validators.rules.amount_ranges import StrategyRoiTargetRanges, T1AmountRanges


class StrategyRoiTargetRule(ValidationRule):
    def __init__(self, goal_type: GoalTypeEnum, roi_target: Optional[Union[float, int]]):
        self.goal_type = goal_type
        self.roi_target = roi_target

    def execute(self):
        """
        In case 'goal_type' == 'roi' the validation rules are:
        - Become required
        - Should be in range from 0.01 to 9999999.99
        """

        if self.goal_type is GoalTypeEnum.ROI and self.roi_target is None:
            raise ValidationException("roi_target is mandatory when goal_type is 'roi'")

        if self.roi_target is None:
            return

        if not is_number(self.roi_target):
            raise ValidationException("roi_target must be a number")

        T1AmountRanges(type_ranges=StrategyRoiTargetRanges,
                       amount=self.roi_target,
                       amount_field="roi_target").execute()



from ts_t1_validator import T1Service
from ts_t1_validator.models.enums.goal_type import GoalTypeEnum
from ts_t1_validator.utils import is_int
from ts_t1_validator.validators.exceptions import ValidationException
from ts_t1_validator.validators.rules.abstract_rule import ValidationRule


class T1MeritPixelRule(ValidationRule):
    def __init__(self, goal_type: GoalTypeEnum, merit_pixel_id: int, t1_service: T1Service):
        self.goal_type = goal_type
        self.merit_pixel_id = merit_pixel_id
        self.t1_service = t1_service

    def execute(self):
        """
        - Must be integer
        - Must be greater than 0
        - Must be existent merit_pixel_id
        - required when goal_type is roi or cpa
        :return:
        """
        if self.goal_type in (GoalTypeEnum.ROI, GoalTypeEnum.CPA) and self.merit_pixel_id is None:
            raise ValidationException("merit_pixel_id is required when goal_type is roi or cpa")

        if self.goal_type not in (GoalTypeEnum.ROI, GoalTypeEnum.CPA) and self.merit_pixel_id is not None:
            raise ValidationException("merit_pixel_id should not be provided when goal_type is not roi or cpa")

        if self.merit_pixel_id is not None:
            # check for attribute type
            if not isinstance(self.merit_pixel_id, (int, float)):
                raise ValidationException("Invalid merit_pixel_id")
            if not is_int(self.merit_pixel_id):
                if self.merit_pixel_id - int(self.merit_pixel_id) > 0:
                    raise ValidationException("merit_pixel_id must be integer")

            merit_pixel_id = int(self.merit_pixel_id)

            if merit_pixel_id < 1:
                raise ValidationException("Invalid merit_pixel_id")

            # check for attribute permission
            if not self.t1_service.isValidMeritPixel(merit_pixel_id):
                raise ValidationException("Invalid merit_pixel_id")



from enum import unique

from ts_t1_validator.models.enums.abstract_enum import AbstractEnum


@unique
class InputTypeEnum(AbstractEnum):
    FIXED = "fixed"
    PERCENTAGE = "percentage"
    UNDEFINED = None



from enum import unique

from ts_t1_validator.models.enums.abstract_enum import AbstractEnum


@unique
class ImpressionPacingIntervalEnum(AbstractEnum):
    HOUR = "hour"
    DAY = "day"
    UNDEFINED = None



from ts_t1_validator.validators.rules.abstract_rule import ValidationRule
from ts_t1_validator.validators.rules.pacing_type import PacingTypeRule
from ts_t1_validator.models.enums.pacing_interval import PacingIntervalEnum
from ts_t1_validator.models.enums.pacing_type import PacingTypeEnum


class PacingTypePutRule(ValidationRule):
    def __init__(self, pacing_type: PacingTypeEnum, pacing_amount: PacingIntervalEnum, pacing_interval):
        assert type(pacing_type) is PacingTypeEnum, "pacing_type should be PacingTypeEnum type"
        assert type(pacing_interval) is PacingIntervalEnum, "pacing_interval should be PacingIntervalEnum type"

        self.pacing_type = pacing_type
        self.pacing_amount = pacing_amount
        self.pacing_interval = pacing_interval

    def execute(self):
        """
        rule:
            pacing_amount, pacing_interval
            if undefined in DB, follow POST logic
        """

        postValidator = PacingTypeRule(pacing_type=self.pacing_type,
                                       pacing_interval=self.pacing_interval,
                                       pacing_amount=self.pacing_amount)

        return postValidator.execute()



from typing import Union

from .abstract_rule import ValidationRule
from ..exceptions import ValidationException
from ts_t1_validator.models.enums.cap_type import CapTypeEnum


class CapAutomaticPutRule(ValidationRule):
    def __init__(self, cap_type: CapTypeEnum, cap_automatic: bool or None,
                 cap_amount: Union[int, float, None], cap_fields=dict) -> None:
        assert type(cap_type) is CapTypeEnum, "cap_type should be type of CapTypeEnum"
        assert type(cap_automatic) is bool or cap_automatic is None, "cap_automatic should be type of bool or None"
        assert isinstance(cap_amount, (int, float, type(None))), "cap_amount should be type of float or None"

        self.cap_type = cap_type
        self.cap_automatic = cap_automatic
        self.cap_amount = cap_amount
        self.cap_fields = cap_fields

    def put(self, cap_automatic: bool, cap_amount: float):
        """
        method to reset fields in case we update the model, there are only 2 cases when we need to reset fields:
        1) user switches type to no-limit, in this case both, amount and automatic should be switched to None
        2) user set amount to True for cap_type ('asap', 'even'), in this case amount should be switched to None

        :param cap_automatic: bool
        :param cap_amount: float
        :return: self
        """

        if self.cap_type is CapTypeEnum.NO_LIMIT:
            self.cap_automatic = cap_automatic
            self.cap_amount = cap_amount
        elif self.cap_type in [CapTypeEnum.ASAP, CapTypeEnum.EVEN] and self.cap_automatic is True:
            self.cap_amount = cap_amount

        return self

    def execute(self) -> None:
        """
        rules:
            1) it can defined alone as TRUE, only if t1_spend_cap_type in DB is asap or even.
            2) it can be defined alone as FALSE only if  DB's t1_spend_cap_amount is already defined.
            3) should be defined for even/asap
        :return:
        """

        # should be defined for even/asap
        if self.cap_automatic is None and self.cap_type in [CapTypeEnum.EVEN, CapTypeEnum.ASAP]:
            raise ValidationException(
                "For %(cap_type)s asap or even %(cap_automatic)s must be True or False" % self.cap_fields)

        # it can defined alone as TRUE, only if t1_spend_cap_type in DB is asap or even.
        if self.cap_automatic in [True, False] and self.cap_type not in [CapTypeEnum.EVEN, CapTypeEnum.ASAP]:
            self.cap_fields["cap_bool"] = str(self.cap_automatic)
            raise ValidationException(
                "For %(cap_automatic)s to be %(cap_bool)s, %(cap_type)s must be asap or even" % self.cap_fields)

        # it can be defined alone as FALSE only if  DB's t1_spend_cap_amount is already defined.
        if self.cap_automatic is False and self.cap_amount is None:
            raise ValidationException("Cannot define %(cap_automatic)s as False without %(cap_amount)s" % self.cap_fields)



from .abstract_rule import ValidationRule
from enum import Enum

from ..exceptions import ValidationException


class CampaignRanges(Enum):
    FIXED_MIN_VALUE = 1.00
    FIXED_MAX_VALUE = 9999999.99
    PERCENTAGE_MIN_VALUE = -99.99
    PERCENTAGE_MAX_VALUE = 9999999.99


class CampaignImpressionRanges(Enum):
    FIXED_MIN_VALUE = 1
    FIXED_MAX_VALUE = 9999999
    PERCENTAGE_MIN_VALUE = -99.99
    PERCENTAGE_MAX_VALUE = 9999999.99


class StrategyRanges(Enum):
    FIXED_MIN_VALUE = 0.01
    FIXED_MAX_VALUE = 9999999.99
    PERCENTAGE_MIN_VALUE = -99.99
    PERCENTAGE_MAX_VALUE = 9999999.99


class StrategyFrequencyRanges(Enum):
    FIXED_MIN_VALUE = 1
    FIXED_MAX_VALUE = 9999999
    PERCENTAGE_MIN_VALUE = -99.99
    PERCENTAGE_MAX_VALUE = 9999999.99


class StrategyImpressionRanges(Enum):
    FIXED_MIN_VALUE = 1
    FIXED_MAX_VALUE = 9999999
    PERCENTAGE_MIN_VALUE = -99.99
    PERCENTAGE_MAX_VALUE = 9999999.99


class StrategyGoalRanges(Enum):
    FIXED_MIN_VALUE = 0.01
    FIXED_MAX_VALUE = 9999999.99
    PERCENTAGE_MIN_VALUE = -99.99
    PERCENTAGE_MAX_VALUE = 9999999.99


class StrategyMaxBidRanges(Enum):
    FIXED_MIN_VALUE = 0.01
    FIXED_MAX_VALUE = 9999999.99
    PERCENTAGE_MIN_VALUE = -99.99
    PERCENTAGE_MAX_VALUE = 9999999.99


class StrategyMinBidRanges(Enum):
    FIXED_MIN_VALUE = 0
    FIXED_MAX_VALUE = 9999999.99
    PERCENTAGE_MIN_VALUE = -99.99
    PERCENTAGE_MAX_VALUE = 9999999.99


class StrategyBudgetRanges(Enum):
    FIXED_MIN_VALUE = 0.01
    FIXED_MAX_VALUE = 9_999_999.99


class StrategyRoiTargetRanges(Enum):
    FIXED_MIN_VALUE = 0.01
    FIXED_MAX_VALUE = 9_999_999.99


class T1AmountRanges(ValidationRule):
    def __init__(self, amount, type_ranges, amount_field):
        assert type(type_ranges) is Enum.__class__

        self.type_ranges = type_ranges
        self.amount = amount
        self.amount_field = amount_field

    def execute(self):
        """
        rules:
            valid range is 0.01 and 9999999.99
        """

        if not (isinstance(self.amount, float) or self.amount is None or isinstance(self.amount, int)):
            raise ValidationException(f"{self.amount_field} must be numeric type")

        template = "value for {field_name} must be between " \
                   "{min_value:,.2f} and {max_value:,.2f}"
        if type(self.type_ranges.FIXED_MIN_VALUE.value) is int:
            template = "value for {field_name} must be between " \
                       "{min_value:,} and {max_value:,}"

        if not (self.type_ranges.FIXED_MIN_VALUE.value <= self.amount <= self.type_ranges.FIXED_MAX_VALUE.value):
            raise ValidationException(template.format(
                **{
                    "field_name": self.amount_field,
                    "min_value": self.type_ranges.FIXED_MIN_VALUE.value,
                    "max_value": self.type_ranges.FIXED_MAX_VALUE.value
                }))



import os

from ts_t1_validator.validators.rules.amount_ranges import T1AmountRanges, CampaignRanges, StrategyFrequencyRanges
from ts_t1_validator.validators.rules.cap_required_fields import CapRequiredFieldsRule
from .abstract_validator import AbstractValidator
from .rules.cap_amount import CapAmountRule
from .rules.cap_automatic import CapAutomaticRule
from .rules.dates_overlap import DatesOverlapRule
from .rules.frequency_required_fields import FrequencyRequiredFieldsRule
from .rules.frequency_type import FrequencyTypeRule
from .rules.goal_type import GoalTypeRule
from .rules.programmatic_guaranteed_overrides import ProgrammaticGuaranteedOverridesRule
from .rules.t1_advertiser import T1AdvertiserRule
from .rules.t1_budget import T1BudgetRule
from .rules.t1_merit_pixel import T1MeritPixelRule
from .. import T1Service
from ..models.dto.campaign import CampaignDTO


class CampaignPostValidator(AbstractValidator):
    def __init__(self, t1_service: T1Service):
        """
        campaign validation manager
        :param t1_service: T1Service
        """
        self.rules = list()
        self.errors = list()
        self.json_schema = os.path.dirname(os.path.abspath(__file__)) + "/campaign/schema/campaign_schema.json"
        self.t1_service = t1_service

    def build_rules_set(self, *args, **kwargs):
        """
        Build rules set
        :param args: Dict
        :param kwargs: Dict
        """
        # remove previous rules
        self.rules = list()

        dto = CampaignDTO.fromDict(kwargs.get("dto", {}))

        # check for advertiser
        if dto.advertiser_id:
            self.rules.append(T1AdvertiserRule(
                advertiser_id=dto.advertiser_id,
                t1_service=self.t1_service))

        # start_date is later than end_date. Message "start_date must be earlier than end_date"
        # start_time is later than end_time. Message " start_time must be earlier than end_time"
        # start_date is in the past. Message "start_date must be in the future"
        # start_date & end_date should be in YYYY-MM-DDTHH:MM:SS format
        if dto.start_date is not None and dto.end_date is not None:
            self.rules.append(DatesOverlapRule(dto.start_date, dto.end_date, dto.zone_name))

        # budget should be converted based on currency code and not be greater than 9,999,999.99.00 in USD
        if dto.budget is not None:
            self.rules.append(T1BudgetRule(
                budget=dto.budget,
                currency=dto.currency_code,
                t1_service=self.t1_service))

        # goal_type should be one of ctr, vcr, viewability_rate, cpa, cpc, reach, roi, spend, vcpm
        # goal_value Must be > 0 and <= 9,999,999.9999 if goal_type in (ROI, CPA, CPC, Viewable
        # CPM, CPM Reach, CPM Spend)
        # goal_value Must be > 0 and <= 100 if goal_type in (CTR, Video Completion Rate (VCR), Viewability Rate (VR))
        if dto.goal_type is not None:
            self.add_rule(GoalTypeRule(goal_type=dto.goal_type,
                                       goal_value=dto.goal_value))

        # pg campaign override values
        if dto.is_programmatic_guaranteed:
            self.rules.append(ProgrammaticGuaranteedOverridesRule(spend_cap_automatic=dto.spend_cap_automatic,
                                                                  spend_cap_type=dto.spend_cap_type,
                                                                  spend_cap_amount=dto.spend_cap_amount,
                                                                  frequency_optimization=dto.frequency_optimization,
                                                                  frequency_type=dto.frequency_type,
                                                                  frequency_interval=dto.frequency_interval,
                                                                  frequency_amount=dto.frequency_amount,
                                                                  restrict_targeting_to_same_device_id=dto.restrict_targeting_to_same_device_id))

        # validate PMP only fields
        else:
            # if present at least one spend cap field, all other should exists as well
            self.rules.append(
                CapRequiredFieldsRule(cap_type=dto.spend_cap_type,
                                      cap_automatic=dto.spend_cap_automatic,
                                      cap_amount=dto.spend_cap_amount,
                                      cap_fields={"cap_automatic": "spend_cap_automatic",
                                                  "cap_type": "spend_cap_type",
                                                  "cap_amount": "spend_cap_amount"}))

            # require spend_cap_automatic if spend_cap_type is not no-limit
            self.rules.append(CapAutomaticRule(cap_type=dto.spend_cap_type,
                                               cap_automatic=dto.spend_cap_automatic,
                                               cap_fields={"cap_automatic": "spend_cap_automatic",
                                                           "cap_type": "spend_cap_type"}))

            # require spend_cap_amount if spend_cap_type is not no-limit and spend_cap_automatic = 0
            self.rules.append(CapAmountRule(
                cap_type=dto.spend_cap_type,
                cap_automatic=dto.spend_cap_automatic,
                cap_amount=dto.spend_cap_amount,
                cap_fields={"cap_automatic": "spend_cap_automatic",
                            "cap_type": "spend_cap_type",
                            "cap_amount": "spend_cap_amount"}))

            # spend_cap_amount should be float >= 1 and <= 9999999.99
            if dto.spend_cap_amount is not None:
                self.rules.append(T1AmountRanges(
                    type_ranges=CampaignRanges,
                    amount=dto.spend_cap_amount,
                    amount_field="spend_cap_amount"))

            # verify frequency fields
            self.rules.append(FrequencyTypeRule(frequency_optimization=dto.frequency_optimization,
                                                frequency_type=dto.frequency_type,
                                                frequency_amount=dto.frequency_amount,
                                                frequency_interval=dto.frequency_interval))

            self.rules.append(FrequencyRequiredFieldsRule(frequency_type=dto.frequency_type,
                                                          frequency_optimization=dto.frequency_optimization,
                                                          frequency_interval=dto.frequency_interval,
                                                          frequency_amount=dto.frequency_amount,
                                                          cap_fields={
                                                              "frequency_optimization": "frequency_optimization",
                                                              "frequency_type": "frequency_type",
                                                              "frequency_amount": "frequency_amount",
                                                              "frequency_interval": "frequency_interval"}))

            # frequency_amount should be type of float >= 1 and <= 9999999
            if dto.frequency_amount is not None:
                self.rules.append(T1AmountRanges(
                    type_ranges=StrategyFrequencyRanges,
                    amount=dto.frequency_amount,
                    amount_field="frequency_amount"))

            self.rules.append(T1MeritPixelRule(
                goal_type=dto.goal_type,
                merit_pixel_id=dto.merit_pixel_id,
                t1_service=self.t1_service))



from .abstract_rule import ValidationRule
from ..exceptions import ValidationException


class T1CheckIds(ValidationRule):
    def __init__(self, include_value_ids, exclude_value_ids):

        self.include_value_ids = include_value_ids
        self.exclude_value_ids = exclude_value_ids

    def execute(self):
        if self.include_value_ids is None or self.exclude_value_ids is None:
            return

        intrsect = set(self.include_value_ids).intersection(self.exclude_value_ids)
        if len(intrsect) > 0:
            raise ValidationException("Cannot have the same value id in both exclude and include lists.")



from ts_t1_validator.utils import is_number
from ts_t1_validator.validators.exceptions import ValidationException
from ts_t1_validator.validators.rules.abstract_rule import ValidationRule
from ts_t1_validator.models.enums.frequency_interval import FrequencyIntervalEnum
from ts_t1_validator.models.enums.frequency_type import FrequencyTypeEnum


class FrequencyTypeRule(ValidationRule):
    def __init__(self,
                 frequency_optimization,
                 frequency_type: FrequencyTypeEnum,
                 frequency_amount,
                 frequency_interval: FrequencyIntervalEnum):
        self.frequency_optimization = frequency_optimization
        self.frequency_type = frequency_type
        self.frequency_amount = frequency_amount
        self.frequency_interval = frequency_interval

    def execute(self):
        if self.frequency_optimization is None:
            if self.frequency_type is not FrequencyTypeEnum.UNDEFINED:
                raise ValidationException("frequency_optimization is required when frequency_type is defined")

            if self.frequency_interval is not FrequencyIntervalEnum.UNDEFINED:
                raise ValidationException("frequency_optimization is required when frequency_interval is defined")

            if self.frequency_amount is not None:
                raise ValidationException("frequency_optimization is required when frequency_amount is defined")

            return

        """
            rules for 'frequency_type':
            required if frequency_optimization is FALSE
                - error message: "{:field_name} is required when frequency_optimization is FALSE"
            must be undefined if frequency_optimization is TRUE
                - error message: "{:field_name} cannot be defined when frequency_optimization is TRUE"
        """
        if self.frequency_optimization is False and self.frequency_type is FrequencyTypeEnum.UNDEFINED:
            raise ValidationException("frequency_type is required when frequency_optimization is 'off'")
        elif self.frequency_optimization is True and self.frequency_type is not FrequencyTypeEnum.UNDEFINED:
            raise ValidationException("frequency_type cannot be defined when frequency_optimization is 'on'")

        """
            rules for 'frequency_amount':
            required if frequency_type is asap or even
                - error message "{:field_name} is required when frequency_type is asap or even"
            must be undefined if frequency_optimization is TRUE
                - error message: "{:field_name} cannot be defined when frequency_optimization is TRUE"
            must be undefined if frequency_type is no-limit.
                - error message: "{:field_name} cannot be defined when frequency_type is no-limit"
        """
        if self.frequency_type in [FrequencyTypeEnum.ASAP, FrequencyTypeEnum.EVEN] and self.frequency_amount is None:
            raise ValidationException("frequency_amount is required when frequency_type is asap or even")
        elif self.frequency_optimization is True and self.frequency_amount is not None:
            raise ValidationException("frequency_amount cannot be defined when frequency_optimization is 'on'")
        elif self.frequency_type is FrequencyTypeEnum.NO_LIMIT and self.frequency_amount is not None:
            raise ValidationException("frequency_amount cannot be defined when frequency_type is no-limit")

        """
            rules for 'frequency_interval':
            required if frequency_type is asap or even
                - error message "{:field_name} is required when frequency_type is asap or even"
            must be undefined if frequency_optimization is TRUE
                - error message: "{:field_name} cannot be defined when frequency_optimization is TRUE"
            must be undefined if frequency_type is no-limit.
                - error message: "{:field_name} cannot be defined when frequency_type is no-limit"
        """
        if self.frequency_type in [FrequencyTypeEnum.ASAP,
                                   FrequencyTypeEnum.EVEN] and self.frequency_interval is FrequencyIntervalEnum.UNDEFINED:
            raise ValidationException("frequency_interval is required when frequency_type is asap or even")
        elif self.frequency_optimization is True and self.frequency_interval is not FrequencyIntervalEnum.UNDEFINED:
            raise ValidationException("frequency_interval cannot be defined when frequency_optimization is 'on'")
        elif self.frequency_type is FrequencyTypeEnum.NO_LIMIT and self.frequency_interval is not FrequencyIntervalEnum.UNDEFINED:
            raise ValidationException("frequency_interval cannot be defined when frequency_type is no-limit")

        if self.frequency_type in [FrequencyTypeEnum.ASAP,
                                   FrequencyTypeEnum.EVEN]:
            if not is_number(self.frequency_amount):
                raise ValidationException("frequency_amount must be a float")



from enum import unique

from ts_t1_validator.models.enums.abstract_enum import AbstractEnum


@unique
class ImpressionPacingTypeEnum(AbstractEnum):
    EVEN = "even"
    ASAP = "asap"
    NO_LIMIT = "no-limit"
    UNDEFINED = None



from ts_t1_validator.validators.exceptions import ValidationException
from ts_t1_validator.validators.rules.abstract_rule import ValidationRule
from ts_t1_validator.models.enums.impression_pacing_type import ImpressionPacingTypeEnum
from ts_t1_validator.models.enums.pacing_type import PacingTypeEnum
from typing import Union, Optional


class StrategyPostRequiredFieldsRule(ValidationRule):
    def __init__(self,
                 pacing_type: PacingTypeEnum,
                 impression_pacing_type: ImpressionPacingTypeEnum,
                 min_bid: Union[float, int, None],
                 max_bid: Union[float, int, None],
                 goal_value: Union[float, int, None],
                 roi_target: Union[float, int, None],
                 frequency_optimization: Optional[bool]):

        self.pacing_type = pacing_type
        self.impression_pacing_type = impression_pacing_type
        self.min_bid = min_bid
        self.max_bid = max_bid
        self.goal_value = goal_value
        self.roi_target = roi_target
        self.frequency_optimization = frequency_optimization

    def execute(self):
        """
        rules:
            at least one of the following is required: pacing_type, impression_pacing_type,
            min_bid, max_bid, goal_value, frequency_optimization,
            roi_target
        :return:
        """

        if self.pacing_type is not PacingTypeEnum.UNDEFINED:
            return

        if self.impression_pacing_type is not ImpressionPacingTypeEnum.UNDEFINED:
            return

        if self.min_bid is not None:
            return

        if self.max_bid is not None:
            return

        if self.goal_value is not None:
            return

        if self.roi_target is not None:
            return

        if self.frequency_optimization is not None:
            return

        raise ValidationException("At least one of the following is required: pacing_type, impression_pacing_type, min_bid, max_bid, goal_value, roi_target, frequency_optimization")



from ts_t1_validator.models.enums.cap_type import CapTypeEnum
from ts_t1_validator.models.enums.frequency_interval import FrequencyIntervalEnum
from ts_t1_validator.models.enums.frequency_type import FrequencyTypeEnum
from ts_t1_validator.models.enums.goal_type import GoalTypeEnum


class CampaignDTO:
    def __init__(self, advertiser_id, campaign_name, is_programmatic_guaranteed, status,
                 spend_cap_automatic, spend_cap_type, spend_cap_amount,
                 zone_name, start_date, end_date, currency_code, budget,
                 goal_type, goal_value, frequency_optimization,
                 frequency_type, frequency_interval, frequency_amount, restrict_targeting_to_same_device_id, merit_pixel_id):

        self.advertiser_id = advertiser_id
        self.campaign_name = campaign_name
        self.is_programmatic_guaranteed = is_programmatic_guaranteed
        self.status = status
        self.spend_cap_automatic = spend_cap_automatic
        self.spend_cap_type = spend_cap_type
        self.spend_cap_amount = spend_cap_amount
        self.zone_name = zone_name
        self.start_date = start_date
        self.end_date = end_date
        self.currency_code = currency_code
        self.budget = budget
        self.goal_type = goal_type
        self.goal_value = goal_value
        self.frequency_optimization = frequency_optimization
        self.frequency_type = frequency_type
        self.frequency_interval = frequency_interval
        self.frequency_amount = frequency_amount
        self.restrict_targeting_to_same_device_id = restrict_targeting_to_same_device_id
        self.merit_pixel_id = merit_pixel_id

    @staticmethod
    def fromDict(user_data: dict):
        """
        transform user data into inner object
        :param user_data: dict
        :return:
        """
        return CampaignDTO(advertiser_id=user_data.get("advertiser_id"),
                           campaign_name=user_data.get("campaign_name"),
                           is_programmatic_guaranteed=user_data.get("is_programmatic_guaranteed"),
                           status=user_data.get("status"),
                           spend_cap_automatic=user_data.get("spend_cap_automatic"),
                           spend_cap_type=CapTypeEnum.set(user_data.get("spend_cap_type")),
                           spend_cap_amount=user_data.get("spend_cap_amount"),
                           zone_name=user_data.get("zone_name"),
                           start_date=user_data.get("start_date"),
                           end_date=user_data.get("end_date"),
                           currency_code=user_data.get("currency_code"),
                           budget=user_data.get("budget"),
                           goal_type=GoalTypeEnum.set(user_data.get("goal_type")),
                           goal_value=user_data.get("goal_value"),
                           frequency_optimization=user_data.get("frequency_optimization"),
                           frequency_type=FrequencyTypeEnum.set(user_data.get("frequency_type")),
                           frequency_interval=FrequencyIntervalEnum.set(user_data.get("frequency_interval")),
                           frequency_amount=user_data.get("frequency_amount"),
                           restrict_targeting_to_same_device_id=user_data.get("restrict_targeting_to_same_device_id"),
                           merit_pixel_id=user_data.get("merit_pixel_id"))



from enum import unique

from ts_t1_validator.models.enums.abstract_enum import AbstractEnum


@unique
class PercentageGoalTypeEnum(AbstractEnum):
    CTR = "ctr"
    VCR = "vcr"
    VR = "viewability_rate"


@unique
class FixedGoalTypeEnum(AbstractEnum):
    CPA = "cpa"
    CPC = "cpc"
    REACH = "reach"
    ROI = "roi"
    SPEND = "spend"
    VCPM = "vcpm"


@unique
class GoalTypeEnum(AbstractEnum):
    CPA = "cpa"
    CPC = "cpc"
    REACH = "reach"
    ROI = "roi"
    SPEND = "spend"
    VCPM = "vcpm"
    CTR = "ctr"
    VCR = "vcr"
    VR = "viewability_rate"
    UNDEFINED = None

    def isFixed(self) -> bool:
        """
        check value in fixed list
        :return: bool
        """
        return self.value in FixedGoalTypeEnum.asList()

    def isPercentage(self) -> bool:
        """
        check value in percentage list
        :return: bool
        """
        return self.value in PercentageGoalTypeEnum.asList()

    @classmethod
    def list(cls) -> str:
        return ", ".join("'{0}'".format(x.value) for x in cls if x.value is not None)



from enum import unique

from ts_t1_validator.models.enums.abstract_enum import AbstractEnum


@unique
class FrequencyIntervalEnum(AbstractEnum):
    HOUR = "hour"
    DAILY = "day"
    WEEKLY = "week"
    MONTH = "month"
    NOTAPPLICABLE = "not-applicable"
    UNDEFINED = None



from enum import Enum, unique


@unique
class SwitcherEnum(Enum):
    ON = "on"
    OFF = "off"
    UNDEFINED = None

    @classmethod
    def set(cls, value):
        return cls(value) if any(value == item.value for item in cls) else cls(None)

    @classmethod
    def fromBool(cls, value):
        if type(value) is bool:
            return cls.set((SwitcherEnum.OFF.value, SwitcherEnum.ON.value)[value])

        return cls.set(value)



from abc import ABC, abstractmethod
from typing import List, Dict
import json

from ts_t1_validator.validators.exceptions import ValidationException, JsonSchemaException
from jsonschema import ValidationError
from jsonschema.validators import Draft7Validator


class AbstractValidator(ABC):
    rules: list
    errors: list
    json_schema: str

    def _loadSchema(self, file_name):
        """
        load schema from json file
        :param file_name:
        :return: string
        """
        with open(file_name) as f:
            data = json.load(f)

        return data

    def jsonSchemaValidation(self, data) -> List:
        """
        run schema validation
        :param data:
        :return: List
        """
        errors = []
        try:
            # load main body
            schema = self._loadSchema(self.json_schema)

            # do primary validation
            v = Draft7Validator(schema)
            # errors = [x.absolute_path[0] + " : " + x.message for x in v.iter_errors(data)]

            for error in v.iter_errors(data):
                try:
                    raise JsonSchemaException(error)
                except JsonSchemaException as e:
                    errors.append(e)

            errors = [str(x) for x in errors]

        except ValidationError as e:
            errors.append(e.absolute_path[0] + " : " + e.message)

        return errors

    def add_rule(self, rule):
        """
        Append rule to common set
        :param rule:
        :return:
        """
        self.rules.append(rule)

    @abstractmethod
    def build_rules_set(self, *args, **kwargs):
        ...

    def validate(self, *args, **kwargs) -> List:
        """
        Run validation
        :return: List
        """
        self.errors = list()
        self.errors.extend(self.jsonSchemaValidation(kwargs.get("dto", {})))
        self.build_rules_set(*args, **kwargs)
        for rule in self.rules:
            try:
                rule.execute()
            except ValidationException as e:
                self.errors.append(str(e))

        return self.errors



from enum import Enum, unique


@unique
class WeeklyRecurringDaysEnum(Enum):
    MONDAY = "monday"
    TUESDAY = "tuesday"
    WEDNESDAY = "wednesday"
    THURSDAY = "thursday"
    FRIDAY = "friday"
    SATURDAY = "saturday"
    SUNDAY = "sunday"
    UNDEFINED = None

    @classmethod
    def set(cls, value):
        return cls(value) if any(value == item.value for item in cls) else cls(None)

    # noinspection PyTypeChecker
    @staticmethod
    def asList():
        return list(WeeklyRecurringDaysEnum)



from enum import unique
from ts_t1_validator.models.enums.abstract_enum import AbstractEnum


@unique
class CapTypeEnum(AbstractEnum):
    EVEN = "even"
    ASAP = "asap"
    NO_LIMIT = "no-limit"
    UNDEFINED = None

    @classmethod
    def list(cls) -> str:
        return ", ".join("'{0}'".format(x.value) for x in cls if x.value is not None)



from ts_t1_validator.validators.exceptions import ValidationException
from ts_t1_validator.validators.rules.abstract_rule import ValidationRule
from ts_t1_validator.models.enums.pacing_interval import PacingIntervalEnum
from ts_t1_validator.models.enums.pacing_type import PacingTypeEnum


class PacingTypeRule(ValidationRule):
    def __init__(self, pacing_type: PacingTypeEnum, pacing_amount, pacing_interval: PacingIntervalEnum):
        self.pacing_type = pacing_type
        self.pacing_amount = pacing_amount
        self.pacing_interval = pacing_interval

    def execute(self):
        if self.pacing_type is not PacingTypeEnum.UNDEFINED:
            """
            rule:
                pacing_amount required if pacing_type is defined
                error message: "{:field_name} is required when pacing_type is defined"
            :return:
            """
            if self.pacing_amount is None:
                raise ValidationException("pacing_amount is required when pacing_type is defined")

            """
            rule:
                pacing_interval required if pacing_type is defined
                error message: "{:field_name} is required when pacing_type is defined"
            :return:
            """
            if self.pacing_interval is PacingIntervalEnum.UNDEFINED:
                raise ValidationException("pacing_interval is required when pacing_type is defined")
        else:
            if self.pacing_interval is not PacingIntervalEnum.UNDEFINED:
                raise ValidationException("pacing_type is required when pacing_interval is defined")

            if self.pacing_amount is not None:
                raise ValidationException("pacing_interval is required when pacing_amount is defined")



from enum import unique
from ts_t1_validator.models.enums.abstract_enum import AbstractEnum


@unique
class StrategyTypeEnum(AbstractEnum):
    REM = "REM"
    GBO = "GBO"
    AUD = "AUD"
    UNDEFINED = None



from enum import unique
from ts_t1_validator.models.enums.abstract_enum import AbstractEnum


@unique
class DimensionActionEnum(AbstractEnum):
    ADD = "add"
    REMOVE = "remove"
    OVERWRITE = "overwrite"
    UNDEFINED = None



from ts_t1_validator import T1Service
from ts_t1_validator.utils import is_int
from ts_t1_validator.validators.exceptions import ValidationException
from ts_t1_validator.validators.rules.abstract_rule import ValidationRule


class T1AdvertiserRule(ValidationRule):
    def __init__(self, advertiser_id: int, t1_service: T1Service):
        self.advertiser_id = advertiser_id
        self.t1_service = t1_service

    def execute(self):
        """
        - Must have T1 permissions
        - Must be integer
        - Must be greater than 0
        :return:
        """
        # check for attribute type
        if not is_int(self.advertiser_id) or self.advertiser_id < 1:
            raise ValidationException("Invalid advertiser_id")

        # check for attribute permission
        if not self.t1_service.isValidAdvertiser(self.advertiser_id):
            raise ValidationException("Invalid advertiser_id")



from .abstract_rule import ValidationRule
from ..exceptions import ValidationException
from ts_t1_validator.models.enums.cap_type import CapTypeEnum
from typing import Union, Optional


class CapAmountPutRule(ValidationRule):
    def __init__(self, cap_type: CapTypeEnum, cap_automatic: Optional[bool], cap_amount: Union[int, float, None],
                 cap_fields=dict):
        assert type(cap_type) is CapTypeEnum, "cap_type should be type of CapTypeEnum"
        assert type(cap_automatic) is bool or cap_automatic is None, "cap_automatic should be type of bool or None"
        assert isinstance(cap_amount, (int, float, type(None))), "cap_amount should be type of float or None"

        self.cap_type = cap_type
        self.cap_automatic = cap_automatic
        self.cap_amount = cap_amount
        self.cap_fields = cap_fields

    def put(self, cap_automatic: bool, cap_amount: float):
        """
        method to reset fields in case we update the model, there are only 2 cases when we need to reset fields:
        1) user switches type to no-limit, in this case both, amount and automatic should be switched to None
        2) user set amount to True for cap_type ('asap', 'even'), in this case amount should be switched to None

        :param cap_automatic: bool
        :param cap_amount: float
        :return: self
        """

        if self.cap_type is CapTypeEnum.NO_LIMIT:
            self.cap_automatic = cap_automatic
            self.cap_amount = cap_amount
        elif self.cap_type in [CapTypeEnum.ASAP, CapTypeEnum.EVEN] and self.cap_automatic is True:
            self.cap_amount = cap_amount

        return self

    def execute(self) -> None:
        """
        rules:
            1) cap_type in asap or even and cap_automatic is false -> cap_amount should not be None
            2) cap_type in asap or even and cap_automatic is true -> cap_amount should be undefined
            2) cap_type is no-limit -> cap_amount should be undefined
        :return:
        """

        # cap_type in asap or even and cap_automatic is false -> cap_amount should not be None
        if self.cap_type in [CapTypeEnum.EVEN, CapTypeEnum.ASAP] and self.cap_automatic is False and self.cap_amount is None:
            raise ValidationException(
                "When %(cap_type)s is even or asap and %(cap_automatic)s is false you must "
                "define %(cap_amount)s." % self.cap_fields)

        # cap_type in asap or even and cap_automatic is true -> cap_amount should be undefined
        if self.cap_type in [CapTypeEnum.EVEN, CapTypeEnum.ASAP] and self.cap_automatic is True and self.cap_amount is not None:
            raise ValidationException("When %(cap_type)s is even or asap and cap %(cap_automatic)s is "
                                      "true %(cap_amount)s must be undefined." % self.cap_fields)

        # cap_type is no-limit -> cap_amount should be undefined
        if self.cap_type is CapTypeEnum.NO_LIMIT and self.cap_amount is not None:
            raise ValidationException("When %(cap_type)s is no-limit %(cap_amount)s "
                                      "must be undefined." % self.cap_fields)



from typing import Optional

from ts_t1_validator.validators.exceptions import ValidationException
from ts_t1_validator.validators.rules.abstract_rule import ValidationRule
from ts_t1_validator.validators.rules.amount_ranges import T1AmountRanges, StrategyMaxBidRanges, StrategyMinBidRanges


class BidsRule(ValidationRule):
    def __init__(self, min_bid: Optional[float], max_bid: Optional[float]):
        self.min_bid = min_bid
        self.max_bid = max_bid

    def execute(self):
        """
        rules:
        if both defined min <= max
        range:  0.01 - 9,999,999.99
        """

        # ignore rule when no bids provided
        if self.min_bid is None and self.max_bid is None:
            return

        if self.min_bid >= self.max_bid:
            raise ValidationException("min_bid must be less than max_bid")

        T1AmountRanges(type_ranges=StrategyMinBidRanges,
                       amount=self.min_bid,
                       amount_field="min_bid").execute()

        T1AmountRanges(type_ranges=StrategyMaxBidRanges,
                       amount=self.max_bid,
                       amount_field="max_bid").execute()



from ts_t1_validator.validators.exceptions import ValidationException
from ts_t1_validator.validators.rules.abstract_rule import ValidationRule


class T1ValueIdsRule(ValidationRule):
    def __init__(self, value_ids, field_name):

        assert type(value_ids) is list or value_ids is None, "value_ids should be a list or None"
        assert type(field_name) is str, "field_name should be a str"

        self.value_ids = value_ids
        self.field_name = field_name

    def execute(self):
        """
        format of provided "ids" is something like: num,num,num

        :return:
        """
        # ignore whole rule if value_ids is None
        if self.value_ids is None:
            return

        # parse str of value_ids into a list
        for value_id in self.value_ids:
            if not (value_id.isdigit() and int(value_id) > 0):
                raise ValidationException("%s should have valid format, positive numbers separated by comma" % self.field_name)



from .abstract_rule import ValidationRule
from ..exceptions import ValidationException
from ts_t1_validator.models.enums.cap_type import CapTypeEnum


class CapRequiredFieldsRule(ValidationRule):
    def __init__(self, cap_type, cap_automatic, cap_amount, cap_fields=dict):
        assert isinstance(cap_type, CapTypeEnum) or cap_type is None

        self.cap_type = cap_type
        self.cap_automatic = cap_automatic
        self.cap_amount = cap_amount
        self.cap_fields = cap_fields

    def put(self, cap_automatic: bool, cap_amount: float):
        """
        method to reset fields in case we update the model, there are only 2 cases when we need to reset fields:
        1) user switches type to no-limit, in this case both, amount and automatic should be switched to None
        2) user set amount to True for cap_type ('asap', 'even'), in this case amount should be switched to None

        :param cap_automatic: bool
        :param cap_amount: float
        :return: self
        """

        if self.cap_type is CapTypeEnum.NO_LIMIT:
            self.cap_automatic = cap_automatic
            self.cap_amount = cap_amount
        elif self.cap_type in [CapTypeEnum.ASAP, CapTypeEnum.EVEN] and self.cap_automatic is True:
            self.cap_amount = cap_amount

        return self

    def execute(self):
        """
        rules:
            if present at least one field, cap_type should be defined as well
        :return:
        """

        if self.cap_amount is not None and self.cap_type is CapTypeEnum.UNDEFINED:
            raise ValidationException(
                "If you define %(cap_automatic)s, %(cap_amount)s you must define %(cap_type)s" % self.cap_fields)

        if self.cap_automatic is not None and self.cap_type is CapTypeEnum.UNDEFINED:
            raise ValidationException(
                "If you define %(cap_automatic)s, %(cap_amount)s you must define %(cap_type)s" % self.cap_fields)



from typing import Dict

import requests as requests


class T1Service:
    def __init__(self, t1_creds: Dict):
        self.__validateCreds(t1_creds)

        self.headers = {"Accept": "application/vnd.mediamath.v1+json",
                        "Authorization": f"Bearer {t1_creds.get('token')}"}
        self.host = t1_creds.get("host")

    def __validateCreds(self, t1_creds: Dict):
        """
        Do simple attribute validation
        :param t1_creds: Dict
        :return:
        """
        if not t1_creds.get("token"):
            raise AttributeError("token attribute was not provided")

        if not t1_creds.get("host"):
            raise AttributeError("host attribute was not provided")

    def getCurrencyRate(self, currency: str) -> float:
        """
        Get currency rate from t1
        :param currency: str
        :return: float
        """

        url = f"{self.host}/currency_rates?full=*&q=currency_code=={currency}&sort_by=-date&page_limit=1"
        conversion_rate = 1
        for _ in range(5):
            try:
                r = requests.get(url, headers=self.headers, timeout=60)
            except requests.ConnectionError:
                continue

            conversion_rate = r.json().get("data", [{}])[0].get("rate", conversion_rate)
            break

        return conversion_rate

    def isValidAdvertiser(self, advertiser_id: int) -> bool:
        """
        Check is advertiser exists
        :param advertiser_id: int
        :return: bool
        """
        url = f"{self.host}/advertisers/{advertiser_id}"
        exists = False
        try:
            r = requests.get(url, headers=self.headers, timeout=60)
            exists = r.status_code == 200
        except requests.ConnectionError:
            ...

        return exists

    def isValidMeritPixel(self, merit_pixel_id: int) -> bool:
        """
        Check is merit pixel exists
        :param merit_pixel_id: int
        :return: bool
        """
        url = f"{self.host}/pixel_bundles/{merit_pixel_id}"
        exists = False
        try:
            r = requests.get(url, headers=self.headers, timeout=60)
            exists = r.status_code == 200
        except requests.ConnectionError:
            ...

        return exists



from ts_t1_validator.models.enums.strategy_type import StrategyTypeEnum
from ts_t1_validator.validators.exceptions import ValidationException
from ts_t1_validator.validators.rules.abstract_rule import ValidationRule


class StrategyTypeRule(ValidationRule):
    def __init__(self, strategy_type: StrategyTypeEnum):
        self.strategy_type = strategy_type

    def execute(self):
        """
        strategy_type should be kind of valid enumeration
        """

        if self.strategy_type is StrategyTypeEnum.UNDEFINED:
            raise ValidationException("strategy_type must be one of " + str(StrategyTypeEnum.asList()))



from enum import unique

from ts_t1_validator.models.enums.abstract_enum import AbstractEnum


@unique
class MediaTypeEnum(AbstractEnum):
    DISPLAY = "display"
    VIDEO = "video"
    UNDEFINED = None



from .abstract_rule import ValidationRule
from ..exceptions import ValidationException


class ImpressionBudgetRule(ValidationRule):
    def __init__(self, impression_cap_automatic: bool or None, total_impression_budget: int or None):
        """
        As a user, I need to be prevented from scheduling a change to automatic campaign impression pacing cap,
        if my existing T1 campaign does not have a defined impression budget

        :param impression_cap_automatic: bool
        :param total_impression_budget: int|None
        """
        assert type(impression_cap_automatic) is bool or impression_cap_automatic is None, \
            "impression_cap_automatic should be bool or None"
        assert type(total_impression_budget) is int or total_impression_budget is None, \
            "total_impression_budget should be int or None"

        self.impression_cap_automatic = impression_cap_automatic
        self.total_impression_budget = total_impression_budget

    def execute(self):
        """
        rules:
        The T1 requirement is that you can only define campaigns.impression_cap_automatic as TRUE if there is
        a campaigns.total_impression_budget of >=1.
        Campaigns.total_impression_budget will never be 0, as T1 requires the value to be >=1.
        It can only be undefined or >=1.
        """

        if self.impression_cap_automatic is True:
            if self.total_impression_budget is None or self.total_impression_budget <= 1:
                raise ValidationException("Cannot set automatic daily impression cap if campaign does not have an impression budget defined")



from enum import Enum


class AbstractEnum(Enum):
    @classmethod
    def set(cls, value):
        return cls(value) if any(value == item.value for item in cls) else cls(None)

    @classmethod
    def asList(cls) -> list:
        """
        Present enum values in list format
        :return: list
        """
        return [x.value for x in list(cls) if x.value is not None]



from ts_t1_validator.models.enums.media_type import MediaTypeEnum
from ts_t1_validator.validators.exceptions import ValidationException
from ts_t1_validator.validators.rules.abstract_rule import ValidationRule


class MediaTypeRule(ValidationRule):
    def __init__(self, media_type: MediaTypeEnum):
        self.media_type = media_type

    def execute(self):
        """
        media_type should be kind of valid enumeration
        """

        if self.media_type is MediaTypeEnum.UNDEFINED:
            raise ValidationException("media_type must be one of " + str(MediaTypeEnum.asList()))



from ts_t1_validator.validators.exceptions import ValidationException
from ts_t1_validator.validators.rules.abstract_rule import ValidationRule
from ts_t1_validator.models.enums.dimension_action import DimensionActionEnum
from ts_t1_validator.models.enums.dimension_code import DimensionCodeEnum


class StrategyDimensionPostRequiredFieldsRule(ValidationRule):
    def __init__(self, dimension_code, action, include_value_ids, exclude_value_ids):

        assert type(dimension_code) is DimensionCodeEnum, "dimension_code should be DimensionCodeEnum type"
        assert type(action) is DimensionActionEnum, "action should be DimensionActionEnum type"

        self.dimension_code = dimension_code
        self.action = action
        self.include_value_ids = include_value_ids
        self.exclude_value_ids = exclude_value_ids

    def execute(self):
        """
        rules:
        - action, dimension_code, include_value_ids or exclude_value_ids are required
        """
        if self.dimension_code is DimensionCodeEnum.UNDEFINED:
            raise ValidationException("dimension_code must be defined in the request")
        if self.action is DimensionActionEnum.UNDEFINED:
            raise ValidationException("action must be defined in the request")

        if self.include_value_ids is None and self.exclude_value_ids is None:
            raise ValidationException("at least one of include_value_ids or exclude_value_ids must be defined")



from typing import Optional

from ts_t1_validator.validators.rules.abstract_rule import ValidationRule
from ts_t1_validator.validators.rules.dates_overlap import DatesOverlapRule


class StrategyDatesRule(ValidationRule):
    campaign_start_date: Optional[str]
    campaign_end_date: Optional[str]
    start_date: Optional[str]
    end_date: Optional[str]
    use_campaign_start_date: bool
    use_campaign_end_date: bool
    zone_name: str

    def __init__(self, use_campaign_start_date: bool, use_campaign_end_date: bool,
                 campaign_start_date: str, campaign_end_date: str,
                 start_date: Optional[str], end_date: Optional[str],
                 zone_name: str):
        self.use_campaign_start_date = use_campaign_start_date
        self.use_campaign_end_date = use_campaign_end_date
        self.campaign_start_date = campaign_start_date
        self.campaign_end_date = campaign_end_date
        self.start_date = start_date
        self.end_date = end_date
        self.zone_name = zone_name

    def execute(self):
        start_date = self.start_date
        end_date = self.end_date
        if self.use_campaign_start_date is True:
            start_date = self.campaign_start_date

        if self.use_campaign_end_date is True:
            end_date = self.campaign_end_date

        DatesOverlapRule(start_date=start_date,
                         end_date=end_date,
                         zone_name=self.zone_name).execute()



from .abstract_rule import ValidationRule
from ..exceptions import ValidationException
from ts_t1_validator.models.enums.cap_type import CapTypeEnum


class CapTypeUndefinedRule(ValidationRule):
    def __init__(self, cap_type1, cap_type2):

        assert isinstance(cap_type1, CapTypeEnum)
        assert isinstance(cap_type2, CapTypeEnum)

        self.cap_type1 = cap_type1
        self.cap_type2 = cap_type2

    def execute(self):
        """
        rule:
            required if impression_cap_type is undefined (one or the other must be defined at a minimum)
        :return:
        """

        if self.cap_type1 is CapTypeEnum.UNDEFINED and self.cap_type2 is CapTypeEnum.UNDEFINED:
            raise ValidationException("At least one of t1_spend_cap_type or impression_cap_type is required")



from .abstract_rule import ValidationRule
from ..exceptions import ValidationException
from ts_t1_validator.models.enums.frequency_type import FrequencyTypeEnum
from ts_t1_validator.models.enums.frequency_interval import FrequencyIntervalEnum


class FrequencyRequiredFieldsRule(ValidationRule):
    def __init__(self, frequency_type: FrequencyTypeEnum, frequency_optimization: bool,
                 frequency_interval: FrequencyIntervalEnum, frequency_amount, cap_fields=dict):
        self.frequency_type = frequency_type
        self.frequency_optimization = frequency_optimization
        self.frequency_amount = frequency_amount
        self.frequency_interval = frequency_interval
        self.cap_fields = cap_fields

    def execute(self):

        if self.frequency_interval is not FrequencyIntervalEnum.UNDEFINED and self.frequency_type is FrequencyTypeEnum.UNDEFINED:
            raise ValidationException(
                "If you define %(frequency_interval)s, you must define %(frequency_type)s" % self.cap_fields)

        if self.frequency_amount is not None and self.frequency_type is FrequencyTypeEnum.UNDEFINED:
            raise ValidationException(
                "If you define %(frequency_amount)s you must define %(frequency_type)s" % self.cap_fields)



import os

from ts_t1_validator.validators.rules.amount_ranges import T1AmountRanges, StrategyFrequencyRanges, StrategyRanges, StrategyImpressionRanges, StrategyGoalRanges
from .abstract_validator import AbstractValidator
from .rules.bids_rule import BidsRule
from .rules.frequency_required_fields import FrequencyRequiredFieldsRule
from .rules.frequency_type import FrequencyTypeRule
from .rules.goal_type import GoalTypeRule
from .rules.impression_pacing_type import ImpressionPacingTypeRule
from .rules.media_type_rule import MediaTypeRule
from .rules.pacing_type import PacingTypeRule
from .rules.roi_target_rule import StrategyRoiTargetRule
from .rules.strategy_dates import StrategyDatesRule
from .rules.strategy_dates_required import StrategyDatesRequiredRule
from .rules.strategy_required_fields_post import StrategyPostRequiredFieldsRule
from .rules.strategy_type_rule import StrategyTypeRule
from .rules.t1_budget import T1BudgetRule
from .. import T1Service
from ..models.dto.campaign import CampaignDTO
from ..models.dto.strategy import StrategyDTO
from ..models.enums.goal_type import GoalTypeEnum


class StrategyPostValidator(AbstractValidator):
    def __init__(self, t1_service: T1Service):
        """
        strategy validation manager
        :param t1_service: T1Service
        """
        self.rules = list()
        self.errors = list()
        self.json_schema = os.path.dirname(os.path.abspath(__file__)) + "/strategy/schema/strategy_schema.json"
        self.t1_service = t1_service

    def build_rules_set(self, *args, **kwargs):
        """
        Build rules set
        """

        # remove previous rules
        self.rules = list()

        dto = StrategyDTO.fromDict(kwargs.get("dto", {}))
        campaign_dto = CampaignDTO.fromDict(kwargs.get("campaign_dto", {}))

        # case: at least one of the following is required: pacing_type, impression_pacing_type,
        # min_bid, max_bid, goal_value, frequency_optimization
        self.rules.append(StrategyPostRequiredFieldsRule(pacing_type=dto.pacing_type,
                                                         impression_pacing_type=dto.impression_pacing_type,
                                                         min_bid=dto.min_bid, max_bid=dto.max_bid,
                                                         goal_value=dto.goal_value,
                                                         roi_target=dto.roi_target,
                                                         frequency_optimization=dto.frequency_optimization))

        # budget should be converted based on currency code and not be greater than 9,999,999.99.00 in USD
        if dto.budget is not None:
            self.rules.append(T1BudgetRule(
                budget=dto.budget,
                currency=campaign_dto.currency_code,
                t1_service=self.t1_service))

        # media_type validation
        self.rules.append(MediaTypeRule(media_type=dto.media_type))

        # strategy_type validation
        self.rules.append(StrategyTypeRule(strategy_type=dto.strategy_type))

        # pacing_* validation
        self.rules.append(PacingTypeRule(pacing_type=dto.pacing_type,
                                         pacing_interval=dto.pacing_interval,
                                         pacing_amount=dto.pacing_amount))

        # impression_pacing_* validation
        self.rules.append(ImpressionPacingTypeRule(impression_pacing_type=dto.impression_pacing_type,
                                                   impression_pacing_amount=dto.impression_pacing_amount,
                                                   impression_pacing_interval=dto.impression_pacing_interval))

        # frequency_* validation
        self.rules.append(FrequencyTypeRule(frequency_optimization=dto.frequency_optimization,
                                            frequency_type=dto.frequency_type,
                                            frequency_amount=dto.frequency_amount,
                                            frequency_interval=dto.frequency_interval))

        # Strategy dates required validator
        self.rules.append(StrategyDatesRequiredRule(use_campaign_start_date=dto.use_campaign_start_date,
                                                    use_campaign_end_date=dto.use_campaign_end_date,
                                                    start_date=dto.start_date,
                                                    end_date=dto.end_date))

        # Strategy dates overlap validator
        self.rules.append(StrategyDatesRule(use_campaign_start_date=dto.use_campaign_start_date,
                                            use_campaign_end_date=dto.use_campaign_end_date,
                                            campaign_start_date=campaign_dto.start_date,
                                            campaign_end_date=campaign_dto.end_date,
                                            start_date=dto.start_date,
                                            end_date=dto.end_date,
                                            zone_name=campaign_dto.zone_name))

        self.rules.append(FrequencyRequiredFieldsRule(frequency_type=dto.frequency_type, frequency_optimization=dto.frequency_optimization,
                                                      frequency_interval=dto.frequency_interval, frequency_amount=dto.frequency_amount,
                                                      cap_fields={"frequency_optimization": "frequency_optimization",
                                                                  "frequency_type": "frequency_type",
                                                                  "frequency_amount": "frequency_amount",
                                                                  "frequency_interval": "frequency_interval"}))

        # goal_type should be one of ctr, vcr, viewability_rate, cpa, cpc, reach, roi, spend, vcpm
        # goal_value Must be > 0 and <= 9,999,999.9999 if goal_type in (ROI, CPA, CPC, Viewable
        # CPM, CPM Reach, CPM Spend)
        # goal_value Must be > 0 and <= 100 if goal_type in (CTR, Video Completion Rate (VCR), Viewability Rate (VR))
        if dto.goal_type is not GoalTypeEnum.UNDEFINED:
            self.add_rule(GoalTypeRule(goal_type=dto.goal_type,
                                       goal_value=dto.goal_value))

        # roi_target rule
        self.rules.append(StrategyRoiTargetRule(goal_type=dto.goal_type, roi_target=dto.roi_target))

        self.rules.append(T1AmountRanges(
            type_ranges=StrategyRanges,
            amount=dto.pacing_amount,
            amount_field="pacing_amount"))

        self.rules.append(T1AmountRanges(
            type_ranges=StrategyImpressionRanges,
            amount=dto.impression_pacing_amount,
            amount_field="impression_pacing_amount"))

        self.rules.append(T1AmountRanges(
            type_ranges=StrategyFrequencyRanges,
            amount=dto.frequency_amount,
            amount_field="frequency_amount"))

        # rule for goal field
        self.rules.append(T1AmountRanges(
            type_ranges=StrategyGoalRanges,
            amount=dto.goal_value,
            amount_field="goal_value"))

        # rule for min/max bids
        self.rules.append(BidsRule(
            min_bid=dto.min_bid,
            max_bid=dto.max_bid
        ))



import unittest

from faker import Faker

MANDATORY_MSG = "%s is required"


class CommonTestCase(unittest.TestCase):
    t1_service = None

    @classmethod
    def setUpClass(cls):
        cls._faker = Faker()

    def setUp(self):
        ...

    def tearDown(self):
        ...

    def check_validator_error(self, *args, **kwargs):
        error = kwargs.get("validator").validate(**kwargs)
        if kwargs.get("expected_error") == "":
            self.assertEqual(len(error), 0, f"Validation error found: {error}")
        else:
            found_error = False
            for e in error:
                if kwargs.get("expected_error") in e:
                    found_error = True

            if found_error is False:
                self.fail(f"{kwargs.get('expected_error')} not found in list of errors")

    def check_rule_error(self, rule, expected_error):
        error = ""
        try:
            rule.execute()
        except Exception as e:
            error = e.args[0]

        self.assertTrue(expected_error in error)



from enum import Enum

from ts_t1_validator import T1Service
from ts_t1_validator.utils import is_int, is_number
from ts_t1_validator.validators.exceptions import ValidationException
from ts_t1_validator.validators.rules.abstract_rule import ValidationRule


class BudgetRanges(Enum):
    MIN_VALUE = 1
    MAX_VALUE = 9_999_999.99


class T1BudgetRule(ValidationRule):
    def __init__(self, budget: float, currency: str, t1_service: T1Service):
        self.budget = budget
        self.currency = currency
        self.t1_service = t1_service

    def execute(self):
        """
        - Must be a float
        - Must be >= 1 and <= 9,999,999.99.00 (in USD - convert from local currency to USD then check)
        :return:
        """
        # we have to ignore rule for invalid data
        if not is_number(self.budget) or not isinstance(self.currency, str):
            return

        self.currency = self.currency.lower()
        # do a conversion in case we receive non usd currency
        rate = 1
        if self.currency != "usd":
            try:
                rate = self.t1_service.getCurrencyRate(self.currency)
            except BaseException:
                ValidationException("Cannot get currency conversion rate")

        # check for ranges
        budget = rate * self.budget
        if BudgetRanges.MIN_VALUE.value > budget or budget > BudgetRanges.MAX_VALUE.value:
            raise ValidationException("budget is out of range")



import os

import setuptools


CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))


def fread(fname):
    with open(os.path.join(CURRENT_DIR, fname)) as f:
        return f.read()


setuptools.setup(
    name="ts-t1-validator",
    version="0.3.4",
    description="Validation logic for TerminalOne models used for api request models",
    long_description=fread('README.md'),
    long_description_content_type='text/markdown',
    license='Apache License, Version 2.0',
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=['python-dotenv',
                      'pytz',
                      'parameterized',
                      'faker',
                      'mock',
                      'requests',
                      'jsonschema'],
    python_requires='>=3.6'
)



from ts_t1_validator.models.enums.goal_type import GoalTypeEnum
from ts_t1_validator.models.enums.input_type import InputTypeEnum
from ts_t1_validator.utils import is_int
from ts_t1_validator.validators.exceptions import ValidationException
from ts_t1_validator.validators.rules.abstract_rule import ValidationRule


class GoalTypeRule(ValidationRule):
    def __init__(self, goal_type: GoalTypeEnum, goal_value):
        self.goal_type = goal_type
        self.goal_value = goal_value
        self.input_type = None
        self.percentage_min_value = 1
        self.percentage_max_value = 100
        self.fixed_min_value = 1
        self.fixed_max_value = 9999999.9999

    def execute(self):
        enum_message = "{0} must be one of [{1}]"

        # goal_type should be one of ctr, vcr, viewability_rate, cpa, cpc, reach, roi, spend, vcpm
        if self.goal_type is GoalTypeEnum.UNDEFINED:
            raise ValidationException(enum_message.format("goal_type", GoalTypeEnum.list()))

        # define type of amount value
        if self.goal_type.isPercentage():
            self.input_type = InputTypeEnum.PERCENTAGE
        else:
            self.input_type = InputTypeEnum.FIXED

        # goal_value should be int or float
        if not (isinstance(self.goal_value, float) or self.goal_value is None or isinstance(self.goal_value, int)):
            raise ValidationException("goal_value must be numeric type")

        # verify goal_value limits based on goal_type
        if self.goal_value is not None:
            if self.input_type is InputTypeEnum.PERCENTAGE:
                if not is_int(self.goal_value):
                    if self.goal_value - int(self.goal_value) > 0:
                        raise ValidationException("goal_value must be integer")
                if not (self.percentage_min_value <= self.goal_value <= self.percentage_max_value):
                    raise ValidationException(
                        "value for {field_name} when input_type=percentage must be between "
                        "{min_value:,.0f} and {max_value:,.0f}".format(
                            **{
                                "field_name": "goal_value",
                                "min_value": self.percentage_min_value,
                                "max_value": self.percentage_max_value
                            }))
            else:

                template = "value for {field_name} when input_type=fixed must be between " \
                           "{min_value:,.4f} and {max_value:,.4f}"
                if type(self.goal_value) is not float:
                    raise ValidationException("goal_value must be float")
                if not (self.fixed_min_value <= self.goal_value <= self.fixed_max_value):
                    raise ValidationException(template.format(**{
                        "field_name": "goal_value",
                        "min_value": self.fixed_min_value,
                        "max_value": self.fixed_max_value
                    }))



from abc import ABC, abstractmethod
from enum import Enum


class ValidationRule(ABC):

    @abstractmethod
    def execute(self):
        ...

    def getUpdatedValue(self, key):
        """
        get update attribute, dto object (user value) has priority

        :param key:
        :return:
        """
        dto_attr = getattr(self, key)

        if isinstance(dto_attr["dto"], Enum):
            if dto_attr["dto"].value is not None:
                return dto_attr["dto"]
        else:
            if dto_attr["dto"] is not None:
                return dto_attr["dto"]

        if isinstance(dto_attr["db"], Enum):
            if dto_attr["db"].value is not None:
                return dto_attr["db"]
        else:
            if dto_attr["db"] is not None:
                return dto_attr["db"]

        return dto_attr["dto"]



from .abstract_rule import ValidationRule
from ..exceptions import ValidationException
from ...models.enums.cap_type import CapTypeEnum
from ...models.enums.frequency_interval import FrequencyIntervalEnum
from ...models.enums.frequency_type import FrequencyTypeEnum


class ProgrammaticGuaranteedOverridesRule(ValidationRule):
    def __init__(self, spend_cap_type, spend_cap_automatic, spend_cap_amount, frequency_optimization,
                 frequency_type, frequency_interval, frequency_amount, restrict_targeting_to_same_device_id
                 ):
        assert type(spend_cap_type) is CapTypeEnum, "spend_cap_type should be CapTypeEnum type"
        assert type(frequency_type) is FrequencyTypeEnum, "frequency_type should be FrequencyTypeEnum type"
        assert type(frequency_interval) is FrequencyIntervalEnum, \
            "frequency_interval should be FrequencyIntervalEnum type"

        self.spend_cap_type = spend_cap_type
        self.spend_cap_automatic = spend_cap_automatic
        self.spend_cap_amount = spend_cap_amount
        self.frequency_optimization = frequency_optimization
        self.frequency_type = frequency_type
        self.frequency_interval = frequency_interval
        self.frequency_amount = frequency_amount
        self.restrict_targeting_to_same_device_id = restrict_targeting_to_same_device_id

    def execute(self):
        """
        rules:
            spend_cap_type must be 'no-limit'
            spend_cap_automatic must be 'off'
            spend_cap_amount must not be provided
            frequency_optimization must be 'off'
            frequency_type must be 'no-limit'
            frequency_interval must be not-applicable
            frequency_amount must not be provided
            restrict_targeting_to_same_device_id must be 'off'

            :return:
        """
        when_pg_message = "%s when campaign type is PG"
        if self.spend_cap_type != CapTypeEnum.NO_LIMIT:
            raise ValidationException(when_pg_message % "spend_cap_type must be 'no-limit'")

        if self.spend_cap_automatic is not False:
            raise ValidationException(when_pg_message % "spend_cap_automatic must be 'off'")

        if self.spend_cap_amount is not None:
            raise ValidationException(when_pg_message % "spend_cap_amount must not be provided")

        if self.frequency_optimization != 0:
            raise ValidationException(when_pg_message % "frequency_optimization must be 'off'")

        if self.frequency_type != FrequencyTypeEnum.NO_LIMIT:
            raise ValidationException(when_pg_message % "frequency_type must be 'no-limit'")

        if self.frequency_interval != FrequencyIntervalEnum.NOTAPPLICABLE:
            raise ValidationException(when_pg_message % " frequency_interval must be not-applicable")

        if self.frequency_amount is not None:
            raise ValidationException(when_pg_message % "frequency_amount must not be provided")

        if self.restrict_targeting_to_same_device_id != 0:
            raise ValidationException(when_pg_message % "restrict_targeting_to_same_device_id must be 'off'")



from .abstract_rule import ValidationRule
from ..exceptions import ValidationException
from ts_t1_validator.models.enums.cap_type import CapTypeEnum


class CapAmountRule(ValidationRule):
    def __init__(self, cap_type, cap_automatic, cap_amount, cap_fields=dict):
        assert isinstance(cap_type, CapTypeEnum)

        self.cap_type = cap_type
        self.cap_amount = cap_amount
        self.cap_automatic = cap_automatic
        self.cap_fields = cap_fields

    def execute(self):
        """
        rules:
            1) must be undefined if "spend_cap_type" == "no-limit". This is because no-limit overrides this setting in T1
            2) must be undefined if "spend_cap_automatic == TRUE OR undefined, because it is meaningless otherwise
            3) required if "spend_cap_automatic == FALSE
            4) must be undefined if spend_cap_type is undefined too
        :return:
        """

        # must be undefined if spend_cap_type is undefined too
        if self.cap_type is CapTypeEnum.UNDEFINED:
            raise ValidationException(
                "%(cap_amount)s must be undefined when %(cap_type)s is not defined too" % self.cap_fields)

        # must be undefined if "spend_cap_type" == "no-limit". This is because no-limit overrides this setting in T1
        if self.cap_type is CapTypeEnum.NO_LIMIT and self.cap_amount is not None:
            raise ValidationException("%(cap_amount)s must be undefined when %(cap_type)s is no-limit" % self.cap_fields)

        # must be undefined if "spend_cap_automatic == TRUE OR undefined, because it is meaningless otherwise
        if self.cap_automatic is True and self.cap_amount is not None:
            raise ValidationException("%(cap_amount)s must be undefined when %(cap_automatic)s is TRUE" % self.cap_fields)

        # required if "spend_cap_automatic == FALSE
        if self.cap_automatic is False and self.cap_amount is None:
            raise ValidationException("%(cap_amount)s is required when %(cap_automatic)s is FALSE" % self.cap_fields)



from ts_t1_validator.utils import is_number
from ts_t1_validator.validators.exceptions import ValidationException
from ts_t1_validator.validators.rules.abstract_rule import ValidationRule
from typing import Union, Optional

from ts_t1_validator.models.enums.cap_type import CapTypeEnum


class ImpressionCapTypeRule(ValidationRule):
    def __init__(self,
                 impression_cap_type: CapTypeEnum,
                 impression_cap_automatic: Optional[bool],
                 impression_cap_amount: Union[int, float, None]):

        assert isinstance(impression_cap_type, CapTypeEnum), "impression_cap_type should be CapTypeEnum type"
        assert isinstance(impression_cap_automatic, (bool, type(None))), "impression_cap_automatic should be bool type"
        assert isinstance(impression_cap_amount, (int, float, type(None))), \
            "impression_cap_amount should be int or float type"

        self.impression_cap_type = impression_cap_type
        self.impression_cap_automatic = impression_cap_automatic
        self.impression_cap_amount = impression_cap_amount

    def put(self, cap_automatic: bool, cap_amount: Union[int, float, None]):
        """
        method to reset fields in case we update the model, there are only 2 cases when we need to reset fields:
        1) user switches type to no-limit, in this case both, amount and automatic should be switched to None
        2) user set amount to True for cap_type ('asap', 'even'), in this case amount should be switched to None

        :param cap_automatic: bool
        :param cap_amount: float
        :return: self
        """

        if self.impression_cap_type is CapTypeEnum.NO_LIMIT:
            self.impression_cap_automatic = cap_automatic
            self.impression_cap_amount = cap_amount
        elif self.impression_cap_type in [CapTypeEnum.ASAP, CapTypeEnum.EVEN] and self.impression_cap_automatic is True:
            self.impression_cap_amount = cap_amount

        return self

    def execute(self) -> None:
        """
        Check for amount type, for fixed case it should be int, for percentage a float
        :return:
        """
        if self.impression_cap_type is CapTypeEnum.UNDEFINED:
            return

        if self.impression_cap_type in [CapTypeEnum.ASAP, CapTypeEnum.EVEN] and self.impression_cap_automatic is False:
            if not is_number(self.impression_cap_amount):
                raise ValidationException("impression_cap_amount must be a float")

        return



import math


def is_number(s):
    try:
        i = float(s)
        if math.isnan(i):
            raise ValueError

        return True
    except (ValueError, TypeError):
        return False


def is_int(v) -> bool:
    """
    we assume that int we can define real integer or numbers with .0 decimals
    :param v: mixed
    :return: bool
    """
    # return type(v) is int or math.modf(v)[0] == 0
    return isinstance(v, int)


def is_bool(v):
    """
    check that value is bool

    :param v:
    :return: bool
    """
    return type(v) is bool or v.lower() in ("true", "false", "1", "0")


def str2bool(v: str):
    """
    convert str to boolean
    :param v: str
    :return: bool
    """
    if type(v) is bool:
        return v is True
    else:
        return v.lower() in ("true", "1")



from typing import Optional, Union

from ts_t1_validator.utils import is_number
from ts_t1_validator.validators.rules.abstract_rule import ValidationRule
from ts_t1_validator.validators.rules.impression_pacing_type import ImpressionPacingTypeRule
from ts_t1_validator.models.enums.impression_pacing_type import ImpressionPacingTypeEnum
from ts_t1_validator.models.enums.pacing_interval import PacingIntervalEnum


class ImpressionPacingTypePutRule(ValidationRule):
    def __init__(self, impression_pacing_type: ImpressionPacingTypeEnum,
                 impression_pacing_amount: Optional[Union[int, float]],
                 impression_pacing_interval: PacingIntervalEnum):

        assert type(impression_pacing_type) is ImpressionPacingTypeEnum, "impression_pacing_type should be type of ImpressionPacingTypeEnum"
        assert type(impression_pacing_interval) is PacingIntervalEnum, "impression_pacing_interval should be type of PacingIntervalEnum"

        self.impression_pacing_type = impression_pacing_type
        self.impression_pacing_amount = impression_pacing_amount
        self.impression_pacing_interval = impression_pacing_interval

    def put(self, pacing_amount: int or float or None, pacing_interval: PacingIntervalEnum):
        """
        method to reset fields in case we update the model, there is the only case:
        1) user switches type to no-limit, in this case both, amount and interval should be switched to None

        :param pacing_amount: int or float or None
        :param pacing_interval: PacingIntervalEnum
        :return: self
        """

        if self.impression_pacing_type is ImpressionPacingTypeEnum.NO_LIMIT:
            self.impression_pacing_amount = pacing_amount
            self.impression_pacing_interval = pacing_interval

        if is_number(pacing_amount):
            self.impression_pacing_amount = int(pacing_amount)

        return self

    def execute(self):
        """
        rule:
        common logic should stay the same
        """

        validator = ImpressionPacingTypeRule(
            impression_pacing_type=self.impression_pacing_type,
            impression_pacing_amount=self.impression_pacing_amount,
            impression_pacing_interval=self.impression_pacing_interval)

        return validator.execute()



from .abstract_rule import ValidationRule
from ..exceptions import ValidationException
from ts_t1_validator.models.enums.cap_type import CapTypeEnum


class CapTypeSwitchPutRule(ValidationRule):
    def __init__(self, cap_type_new, cap_type_old, cap_fields=dict):

        assert type(cap_type_new) is CapTypeEnum, "cap_type_new should have type of CapTypeEnum"
        assert type(cap_type_old) is CapTypeEnum, "cap_type_new should have type of CapTypeEnum"

        self.cap_type_new = cap_type_new
        self.cap_type_old = cap_type_old
        self.cap_fields = cap_fields

    def execute(self):
        """
        rule:
            If the DB has asap it can be switched to even
            If the DB has even it can be switched to asap
        :return:
        """

        # If the DB has asap it can be switched to even
        if self.cap_type_old is CapTypeEnum.ASAP and self.cap_type_new not in [CapTypeEnum.EVEN, CapTypeEnum.ASAP]:
            raise ValidationException("%(cap_type)s cap_fields must be even" % self.cap_fields)

        # If the DB has even it can be switched to asap
        if self.cap_type_old is CapTypeEnum.EVEN and self.cap_type_new not in [CapTypeEnum.EVEN, CapTypeEnum.ASAP]:
            raise ValidationException("%(cap_type)s cap_fields must be even" % self.cap_fields)



from .abstract_rule import ValidationRule
from ts_t1_validator.models.enums.frequency_type import FrequencyTypeEnum
from ts_t1_validator.models.enums.frequency_interval import FrequencyIntervalEnum
from ts_t1_validator.validators.rules.frequency_required_fields import FrequencyRequiredFieldsRule


class FrequencyRequiredFieldsPutRule(ValidationRule):
    def __init__(self, frequency_type, frequency_optimization, frequency_interval, frequency_amount, cap_fields={}):
        assert isinstance(frequency_type, FrequencyTypeEnum), "frequency_type should be type of FrequencyTypeEnum"
        assert isinstance(frequency_optimization, bool) or frequency_optimization is None, \
            "frequency_optimization should be type of bool or None"
        assert isinstance(frequency_interval, FrequencyIntervalEnum), \
            "frequency_interval should be type of FrequencyIntervalEnum"

        self.frequency_type = frequency_type
        self.frequency_optimization = frequency_optimization
        self.frequency_amount = frequency_amount
        self.frequency_interval = frequency_interval
        self.cap_fields = cap_fields

    def put(self, frequency_amount: float or None, frequency_interval: FrequencyIntervalEnum, frequency_type: FrequencyTypeEnum):
        """
        method to reset fields in case we update the model, there is the only case:
        1) user switches type to no-limit, in this case both, amount and interval should be switched to None

        :param frequency_amount: float or None
        :param frequency_interval: FrequencyIntervalEnum
        :param frequency_type: FrequencyTypeEnum
        :return: self
        """

        if self.frequency_type is FrequencyTypeEnum.NO_LIMIT:
            self.frequency_interval = frequency_interval
            self.frequency_amount = frequency_amount

        if self.frequency_optimization:
            self.frequency_interval = frequency_interval
            self.frequency_amount = frequency_amount
            self.frequency_type = frequency_type

        return self

    def execute(self):
        """
        rule:
        common logic should stay the same
        """

        postValidator = (FrequencyRequiredFieldsRule(frequency_type=self.frequency_type, frequency_optimization=self.frequency_optimization,
                                                     frequency_interval=self.frequency_interval, frequency_amount=self.frequency_amount,
                                                     cap_fields={"frequency_optimization": "frequency_optimization",
                                                                 "frequency_type": "frequency_type",
                                                                 "frequency_amount": "frequency_amount",
                                                                 "frequency_interval": "frequency_interval"}))

        return postValidator.execute()



from typing import Union, Optional

from ts_t1_validator.models.enums.frequency_interval import FrequencyIntervalEnum
from ts_t1_validator.models.enums.frequency_type import FrequencyTypeEnum
from ts_t1_validator.models.enums.goal_type import GoalTypeEnum
from ts_t1_validator.models.enums.impression_pacing_interval import ImpressionPacingIntervalEnum
from ts_t1_validator.models.enums.impression_pacing_type import ImpressionPacingTypeEnum
from ts_t1_validator.models.enums.media_type import MediaTypeEnum
from ts_t1_validator.models.enums.pacing_interval import PacingIntervalEnum
from ts_t1_validator.models.enums.pacing_type import PacingTypeEnum
from ts_t1_validator.models.enums.strategy_type import StrategyTypeEnum


class StrategyDTO:
    campaign_id: int
    name: str
    status: bool
    media_type: str
    strategy_type: str
    use_campaign_start_date: bool
    use_campaign_end_date: bool
    start_date: str
    end_date: str
    budget: Optional[Union[int, float]]
    use_optimization: bool
    goal_type: str
    bid_price_is_media_only: bool
    cost_passthrough_cpm_enabled: bool
    cost_passthrough_percent_enabled: bool

    def __init__(self, campaign_id, name, status, media_type, strategy_type,
                 use_campaign_start_date, use_campaign_end_date,
                 start_date, end_date, budget,
                 pacing_amount, pacing_interval, pacing_type,
                 impression_pacing_amount, impression_pacing_type,
                 impression_pacing_interval, bid_price_is_media_only,
                 min_bid, max_bid,
                 frequency_type, frequency_amount, frequency_interval, frequency_optimization,
                 use_optimization,
                 goal_type, goal_value, roi_target, cost_passthrough_cpm_enabled,
                 cost_passthrough_percent_enabled):

        self.campaign_id = campaign_id
        self.name = name
        self.status = status
        self.media_type = media_type
        self.strategy_type = strategy_type
        self.use_campaign_start_date = use_campaign_start_date
        self.use_campaign_end_date = use_campaign_end_date
        self.start_date = start_date
        self.end_date = end_date
        self.budget = budget
        self.pacing_amount = pacing_amount
        self.pacing_interval = pacing_interval
        self.pacing_type = pacing_type
        self.impression_pacing_amount = impression_pacing_amount
        self.impression_pacing_interval = impression_pacing_interval
        self.impression_pacing_type = impression_pacing_type
        self.bid_price_is_media_only = bid_price_is_media_only
        self.min_bid = min_bid
        self.max_bid = max_bid
        self.frequency_type = frequency_type
        self.frequency_amount = frequency_amount
        self.frequency_interval = frequency_interval
        self.frequency_optimization = frequency_optimization
        self.use_optimization = use_optimization
        self.goal_type = goal_type
        self.goal_value = goal_value
        self.roi_target = roi_target
        self.cost_passthrough_cpm_enabled = cost_passthrough_cpm_enabled
        self.cost_passthrough_percent_enabled = cost_passthrough_percent_enabled

    @staticmethod
    def fromDict(user_data: dict):
        """
        transform user data into inner object
        :param user_data: dict
        :return:
        """
        return StrategyDTO(campaign_id=user_data.get("campaign_id"),
                           name=user_data.get("name"),
                           status=user_data.get("status"),
                           media_type=MediaTypeEnum.set(user_data.get("media_type")),
                           strategy_type=StrategyTypeEnum.set(user_data.get("strategy_type")),
                           use_campaign_start_date=user_data.get("use_campaign_start_date"),
                           use_campaign_end_date=user_data.get("use_campaign_end_date"),
                           start_date=user_data.get("start_date"),
                           end_date=user_data.get("end_date"),
                           budget=user_data.get("budget"),
                           pacing_amount=user_data.get("pacing_amount"),
                           pacing_interval=PacingIntervalEnum.set(user_data.get("pacing_interval")),
                           pacing_type=PacingTypeEnum.set(user_data.get("pacing_type")),
                           impression_pacing_amount=user_data.get("impression_pacing_amount"),
                           impression_pacing_type=ImpressionPacingTypeEnum.set(user_data.get("impression_pacing_type")),
                           impression_pacing_interval=ImpressionPacingIntervalEnum.set(user_data.get("impression_pacing_interval")),
                           bid_price_is_media_only=user_data.get("bid_price_is_media_only"),
                           min_bid=user_data.get("min_bid"),
                           max_bid=user_data.get("max_bid"),
                           frequency_type=FrequencyTypeEnum.set(user_data.get("frequency_type")),
                           frequency_amount=user_data.get("frequency_amount"),
                           frequency_interval=FrequencyIntervalEnum.set(user_data.get("frequency_interval")),
                           frequency_optimization=user_data.get("frequency_optimization"),
                           use_optimization=user_data.get("use_optimization"),
                           goal_type=GoalTypeEnum.set(user_data.get("goal_type")),
                           goal_value=user_data.get("goal_value"),
                           roi_target=user_data.get("roi_target"),
                           cost_passthrough_cpm_enabled=user_data.get("cost_passthrough_cpm_enabled"),
                           cost_passthrough_percent_enabled=user_data.get("cost_passthrough_percent_enabled"))



from typing import Optional

from ts_t1_validator.validators.exceptions import ValidationException
from ts_t1_validator.validators.rules.abstract_rule import ValidationRule


class StrategyDatesRequiredRule(ValidationRule):
    start_date: Optional[str]
    end_date: Optional[str]
    use_campaign_start_date: bool
    use_campaign_end_date: bool

    def __init__(self, use_campaign_start_date: bool, use_campaign_end_date: bool,
                 start_date: Optional[str], end_date: Optional[str]):
        self.use_campaign_start_date = use_campaign_start_date
        self.use_campaign_end_date = use_campaign_end_date
        self.start_date = start_date
        self.end_date = end_date

    def execute(self):
        if self.use_campaign_start_date is False and not self.start_date:
            raise ValidationException("start_date is required when use_campaign_start_date is False")

        if self.use_campaign_end_date is False and not self.end_date:
            raise ValidationException("end_date is required when use_campaign_end_date is False")



from enum import unique

from ts_t1_validator.models.enums.abstract_enum import AbstractEnum


@unique
class RecurrenceTypeEnum(AbstractEnum):
    DAILY = "daily"
    WEEKLY = "weekly"
    MONTHLY = "monthly"
    UNDEFINED = None



from .abstract_rule import ValidationRule
from ..exceptions import ValidationException
from ts_t1_validator.models.enums.cap_type import CapTypeEnum


class CapAutomaticRule(ValidationRule):
    def __init__(self, cap_type, cap_automatic, cap_fields=dict):
        assert isinstance(cap_type, CapTypeEnum)

        self.cap_type = cap_type
        self.cap_automatic = cap_automatic
        self.cap_fields = cap_fields

    def execute(self):
        """
        rules:
            1) must be undefined if "spend_cap_type" == "no-limit". This is because no-limit overrides this setting in T1
            2) required if "spend_cap_type" != "no-limit", but is defined
            3) must be undefined if spend_cap_type is undefined too
        :return:
        """

        if self.cap_type is CapTypeEnum.UNDEFINED or self.cap_type is None:
            raise ValidationException(
                "%(cap_type)s is required field" % self.cap_fields)

        if self.cap_type is CapTypeEnum.NO_LIMIT and self.cap_automatic is not None:
            raise ValidationException("%(cap_automatic)s must be undefined when %(cap_type)s is no-limit" % self.cap_fields)

        if self.cap_type in [CapTypeEnum.EVEN, CapTypeEnum.ASAP] and self.cap_automatic is None:
            raise ValidationException("%(cap_automatic)s is required when %(cap_type)s is asap or even" % self.cap_fields)
